@@ -1,4 +1,4 @@
-// This library is part of PLINK 2.0, copyright (C) 2005-2025 Shaun Purcell,
+// This library is part of PLINK 2.0, copyright (C) 2005-2026 Shaun Purcell,
 // Christopher Chang.
 //
 // This library is free software: you can redistribute it and/or modify it
@@ -30,17 +30,6 @@ const char kErrprintfDecompress[] = "Error: %s decompression failure: %s.\n";
 
 uint64_t g_failed_alloc_attempt_size = 0;
 
-#if (((__GNUC__ == 4) && (__GNUC_MINOR__ < 7)) || (__GNUC__ >= 11)) && !defined(__APPLE__)
-BoolErr pgl_malloc(uintptr_t size, void* pp) {
-  *S_CAST(unsigned char**, pp) = S_CAST(unsigned char*, malloc(size));
-  if (likely(*S_CAST(unsigned char**, pp))) {
-    return 0;
-  }
-  g_failed_alloc_attempt_size = size;
-  return 1;
-}
-#endif
-
 BoolErr fwrite_checked(const void* buf, uintptr_t len, FILE* outfile) {
   while (len > kMaxBytesPerIO) {
     // OS X fwrite() doesn't support 2GiB+ writes
@@ -366,7 +355,7 @@ int32_t memequal(const void* m1, const void* m2, uintptr_t byte_ct) {
     // tried unrolling this, doesn't make a difference
     const VecUc v1 = vecuc_loadu(&(m1_alias[vidx]));
     const VecUc v2 = vecuc_loadu(&(m2_alias[vidx]));
-    if (vecuc_movemask(v1 == v2) != kVec8thUintMax) {
+    if (!vecucs_are_equal(v1, v2)) {
       return 0;
     }
   }
@@ -376,7 +365,7 @@ int32_t memequal(const void* m1, const void* m2, uintptr_t byte_ct) {
     const uintptr_t final_offset = byte_ct - kBytesPerVec;
     const VecUc v1 = vecuc_loadu(&(m1_uc[final_offset]));
     const VecUc v2 = vecuc_loadu(&(m2_uc[final_offset]));
-    if (vecuc_movemask(v1 == v2) != kVec8thUintMax) {
+    if (!vecucs_are_equal(v1, v2)) {
       return 0;
     }
   }
@@ -460,22 +449,39 @@ int32_t Memcmp(const void* m1, const void* m2, uintptr_t byte_ct) {
   for (uintptr_t vidx = 0; vidx != fullvec_ct; ++vidx) {
     const VecUc v1 = vecuc_loadu(&(m1_alias[vidx]));
     const VecUc v2 = vecuc_loadu(&(m2_alias[vidx]));
+#  ifndef SIMDE_ARM_NEON_A32V8_NATIVE
     // is this even worthwhile now in non-AVX2 case?
     const uint32_t movemask_result = vecuc_movemask(v1 == v2);
     if (movemask_result != kVec8thUintMax) {
+      // todo: check if this is faster to do outside of vector-space
       const uintptr_t diff_pos = vidx * kBytesPerVec + ctzu32(~movemask_result);
       return (m1_uc[diff_pos] < m2_uc[diff_pos])? -1 : 1;
     }
+#  else
+    const uint64_t eq_nybbles = arm_shrn4_uc(v1 == v2);
+    if (eq_nybbles != UINT64_MAX) {
+      const uintptr_t diff_pos = vidx * kBytesPerVec + ctzw(~eq_nybbles) / 4;
+      return (m1_uc[diff_pos] < m2_uc[diff_pos])? -1 : 1;
+    }
+#  endif
   }
   if (byte_ct % kBytesPerVec) {
     const uintptr_t final_offset = byte_ct - kBytesPerVec;
     const VecUc v1 = vecuc_loadu(&(m1_uc[final_offset]));
     const VecUc v2 = vecuc_loadu(&(m2_uc[final_offset]));
+#  ifndef SIMDE_ARM_NEON_A32V8_NATIVE
     const uint32_t movemask_result = vecuc_movemask(v1 == v2);
     if (movemask_result != kVec8thUintMax) {
       const uintptr_t diff_pos = final_offset + ctzu32(~movemask_result);
       return (m1_uc[diff_pos] < m2_uc[diff_pos])? -1 : 1;
     }
+#  else
+    const uint64_t eq_nybbles = arm_shrn4_uc(v1 == v2);
+    if (eq_nybbles != UINT64_MAX) {
+      const uintptr_t diff_pos = final_offset + ctzw(~eq_nybbles) / 4;
+      return (m1_uc[diff_pos] < m2_uc[diff_pos])? -1 : 1;
+    }
+#  endif
   }
   return 0;
 }
@@ -594,6 +600,43 @@ char* i64toa(int64_t llii, char* start) {
   return uitoa_z8(bottom_eight, start);
 }
 
+uint32_t MaxElementU32(const uint32_t* u32arr, uintptr_t entry_ct) {
+  // confirmed that *std::max_element() has horrible performance on at least
+  // macOS
+  // also confirmed that macOS compiler autovectorizes this; main loop
+  // processes 4 vectors at a time
+  uint32_t result = u32arr[0];
+  for (uintptr_t entry_idx = 1; entry_idx != entry_ct; ++entry_idx) {
+    const uint32_t cur_element = u32arr[entry_idx];
+    if (cur_element > result) {
+      result = cur_element;
+    }
+  }
+  return result;
+}
+
+double MaxElementD(const double* darr, uintptr_t entry_ct) {
+  double result = darr[0];
+  for (uintptr_t entry_idx = 1; entry_idx != entry_ct; ++entry_idx) {
+    const double cur_element = darr[entry_idx];
+    if (cur_element > result) {
+      result = cur_element;
+    }
+  }
+  return result;
+}
+
+double MinElementD(const double* darr, uintptr_t entry_ct) {
+  double result = darr[0];
+  for (uintptr_t entry_idx = 1; entry_idx != entry_ct; ++entry_idx) {
+    const double cur_element = darr[entry_idx];
+    if (cur_element < result) {
+      result = cur_element;
+    }
+  }
+  return result;
+}
+
 #if defined(USE_SSE2) && !defined(NO_UNALIGNED)
 uintptr_t FirstUnequal4(const void* arr1, const void* arr2, uintptr_t nbytes) {
   // Similar to memequal().
@@ -640,21 +683,37 @@ uintptr_t FirstUnequal4(const void* arr1, const void* arr2, uintptr_t nbytes) {
   for (uintptr_t vidx = 0; vidx != vec_ct; ++vidx) {
     const VecUc v1 = vecuc_loadu(&(arr1_alias[vidx]));
     const VecUc v2 = vecuc_loadu(&(arr2_alias[vidx]));
-    const uint32_t eq_result = vecw_movemask(v1 == v2);
+#  ifndef SIMDE_ARM_NEON_A32V8_NATIVE
+    const uint32_t eq_result = vecuc_movemask(v1 == v2);
     if (eq_result != kVec8thUintMax) {
       return vidx * kBytesPerVec + ctzu32(~eq_result);
     }
+#  else
+    const uint64_t eq_nybbles = arm_shrn4_uc(v1 == v2);
+    if (eq_nybbles != UINT64_MAX) {
+      return vidx * kBytesPerVec + ctzw(~eq_nybbles) / 4;
+    }
+#  endif
   }
   if (nbytes % kBytesPerVec) {
     const uintptr_t final_offset = nbytes - kBytesPerVec;
     const char* s1 = S_CAST(const char*, arr1);
     const char* s2 = S_CAST(const char*, arr2);
-    const VecW v1 = vecw_loadu(&(s1[final_offset]));
-    const VecW v2 = vecw_loadu(&(s2[final_offset]));
-    const uint32_t eq_result = vecw_movemask(v1 == v2);
+    // bugfix (5 Jul 2025): this must be VecUc on ARM, not VecW
+    // (disturbing that arm_shrn4_uc() call compiled at all...)
+    const VecUc v1 = vecuc_loadu(&(s1[final_offset]));
+    const VecUc v2 = vecuc_loadu(&(s2[final_offset]));
+#  ifndef SIMDE_ARM_NEON_A32V8_NATIVE
+    const uint32_t eq_result = vecuc_movemask(v1 == v2);
     if (eq_result != kVec8thUintMax) {
       return final_offset + ctzu32(~eq_result);
     }
+#  else
+    const uint64_t eq_nybbles = arm_shrn4_uc(v1 == v2);
+    if (eq_nybbles != UINT64_MAX) {
+      return final_offset + ctzw(~eq_nybbles) / 4;
+    }
+#  endif
   }
   return nbytes;
 }
@@ -695,6 +754,7 @@ uintptr_t CountVintsNonempty(const unsigned char* buf, const unsigned char* buf_
   const uintptr_t ending_addr = R_CAST(uintptr_t, buf_end);
   const VecUc* buf_vlast = R_CAST(const VecUc*, RoundDownPow2(ending_addr - 1, kBytesPerVec));
   const uint32_t leading_byte_ct = starting_addr - R_CAST(uintptr_t, buf_viter);
+  // todo: better ARM implementation
   Vec8thUint vint_ends = (UINT32_MAX << leading_byte_ct) & (~vecuc_movemask(*buf_viter));
   uintptr_t total = 0;
   while (buf_viter != buf_vlast) {
